#!/usr/bin/env python

"""
whisker/socket.py

===============================================================================

    Copyright © 2011-2020 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of the Whisker Python client library.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Low-level network socket functions.**

"""

import re
import socket
from typing import Union

from whisker.constants import BUFFERSIZE


def get_port(x: Union[str, int]) -> int:
    """
    Works out an integer TCP/IP port number.

    Args:
        x: port number or name

    Returns:
        port number

    Raises:
        ValueError: bad value
        TypeError: bad type

    """
    if type(x) is int:
        return x
    m = re.match(r"\D", x)  # search for \D = non-digit characters
    if m:
        port = socket.getservbyname(x, "tcp")
    else:
        port = int(x)
    return port


# In Python 3, we work with strings within the client code, and bytes
# to/from the socket. Translation occurs here:


def socket_receive(sock: socket.socket, bufsize: int = BUFFERSIZE) -> str:
    """
    Receives ASCII data from a socket and returns a string.

    Args:
        sock: TCP/IP socket
        bufsize: buffer size

    Returns:
        data as a string
    """
    # return socket.recv(bufsize)  # Python 2
    return sock.recv(bufsize).decode('ascii')  # Python 3


def socket_sendall(sock: socket.socket, data: str) -> None:
    """
    Sends all the data specified to a network socket, encoded via ASCII.

    See https://stackoverflow.com/questions/34252273.
    """
    # return socket.sendall(data)  # Python 2
    return sock.sendall(data.encode('ascii'))  # Python 3


def socket_send(sock: socket.socket, data: str) -> int:
    """
    Sends some of the data to a network socket, encoded via ASCII,
    and returns the number of bytes actually sent.

    See https://stackoverflow.com/questions/34252273.
    """
    # return socket.send(data)  # Python 2
    return sock.send(data.encode('ascii'))  # Python 3



#!/usr/bin/env python

"""
whisker/qtclient.py

===============================================================================

    Copyright © 2011-2020 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of the Whisker Python client library.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Multithreaded framework for Whisker Python clients using Qt.**

- Created: late 2016
- Last update: 20 Sep 2018

- Note funny bug: data sometimes sent twice.

  - Looks like it might be this:
    http://www.qtcentre.org/threads/29462-QTcpSocket-sends-data-twice-with-flush()
    
  - Attempted solution:
  
    - change ``QTcpSocket()`` to ``QTcpSocket(parent=self)``, in case the
      socket wasn't getting moved between threads properly -- didn't fix
    - disable ``flush()`` -- didn't fix.
    - send function is only being called once, according to log
    - adding thread ID information to the log shows that ``whisker_controller``
      events are coming from two threads...
      
  - Anyway, bug was this:
  
    - http://stackoverflow.com/questions/34125065
    - https://bugreports.qt.io/browse/PYSIDE-249

- Source:

  - http://code.qt.io/cgit/qt/qtbase.git/tree/src/corelib/kernel/qobject.h?h=5.4
  - http://code.qt.io/cgit/qt/qtbase.git/tree/src/corelib/kernel/qobject.cpp?h=5.4
  
"""  # noqa

import logging
from enum import Enum
from typing import Optional

import arrow
from cardinal_pythonlib.regexfunc import CompiledRegexMemory
# noinspection PyPackageRequirements
from PyQt5.QtCore import (
    QByteArray,
    QObject,
    Qt,
    QThread,
    pyqtSignal,
    pyqtSlot,
)
# noinspection PyPackageRequirements
from PyQt5.QtNetwork import (
    QAbstractSocket,
    QTcpSocket,
)

from whisker.api import (
    CLIENT_MESSAGE_REGEX,
    CODE_REGEX,
    ENCODING,
    EOL,
    EOL_LEN,
    ERROR_REGEX,
    EVENT_REGEX,
    IMMPORT_REGEX,
    KEY_EVENT_REGEX,
    msg_from_args,
    PING,
    PING_ACK,
    split_timestamp,
    SYNTAX_ERROR_REGEX,
    WARNING_REGEX,
    WhiskerApi,
)
from whisker.constants import DEFAULT_PORT
from whisker.qt import exit_on_exception, StatusMixin

log = logging.getLogger(__name__)

INFINITE_WAIT = -1


class ThreadOwnerState(Enum):
    STOPPED = 0
    STARTING = 1
    RUNNING = 2
    STOPPING = 3


def is_socket_connected(socket: QAbstractSocket) -> bool:
    """
    Is the Qt socket connected?
    """
    return socket and socket.state() == QAbstractSocket.ConnectedState


def disable_nagle(socket: QAbstractSocket) -> None:
    """
    Disable the Nagle algorithm on the Qt socket. (This makes it send any
    outbound data immediately, rather than waiting and packaging it up with
    potential subsequent data. So this function chooses a low-latency mode
    over a high-efficiency mode.)
    """
    # http://doc.qt.io/qt-5/qabstractsocket.html#SocketOption-enum
    socket.setSocketOption(QAbstractSocket.LowDelayOption, 1)


def get_socket_error(socket: QAbstractSocket) -> str:
    """
    Returns a textual description of the last error on the specified Qt
    socket.
    """
    return "{}: {}".format(socket.error(), socket.errorString())


def quote(msg: str) -> str:
    """
    Return its argument suitably quoted for transmission to Whisker.

    - Whisker has quite a primitive quoting system...
    - Check with strings that actually include quotes.
    """
    return '"{}"'.format(msg)


# =============================================================================
# Object to supervise all Whisker functions
# =============================================================================

class WhiskerOwner(QObject, StatusMixin):  # GUI thread
    """
    Object to own and manage communication with a Whisker server.

    This object is owned by the GUI thread.
    It devolves work to two other threads:

    (A) **main socket listener** (:class:`WhiskerMainSocketListener`, as the
        instance ``self.mainsock``, in the thread ``self.mainsockthread``);
        
    (B) **task** (:class:`WhiskerQtTask`, as the instance ``self.task`` in the
        thread ``self.taskthread``) + **immediate socket blocking handler**
        (:class:`WhiskerController`, as the instance ``self.controller`` in the
        thread ``self.taskthread``).

    References to "main" here just refers to the main socket (as opposed to the
    immediate socket), not the thread that's doing most of the processing.

    Signals of relevance to users:

    - ``connected()``
    - ``disconnected()``
    - ``finished()``
    - ``message_received(str, arrow.Arrow, int)``
    - ``event_received(str, arrow.Arrow, int)``
    - ``pingack_received(arrow.Arrow, int)``
    """
    # Outwards, to world/task:
    connected = pyqtSignal()
    disconnected = pyqtSignal()
    finished = pyqtSignal()
    message_received = pyqtSignal(str, arrow.Arrow, int)
    event_received = pyqtSignal(str, arrow.Arrow, int)
    pingack_received = pyqtSignal(arrow.Arrow, int)
    # Inwards, to possessions:
    controller_finish_requested = pyqtSignal()
    mainsock_finish_requested = pyqtSignal()
    ping_requested = pyqtSignal()
    # And don't forget the signals inherited from StatusMixin.

    # noinspection PyUnresolvedReferences
    def __init__(self,
                 task: 'WhiskerQtTask',  # forward reference for type hint
                 server: str,
                 main_port: int = DEFAULT_PORT,
                 parent: QObject = None,
                 connect_timeout_ms: int = 5000,
                 read_timeout_ms: int = 500,
                 name: str = "whisker_owner",
                 sysevent_prefix: str = 'sys_',
                 **kwargs) -> None:
        """
        Args:
            task: instance of something derived from :class:`WhiskerQtTask`
            server: host name or IP address of Whisker server
            main_port: main port number for Whisker server
            parent: optional parent :`QObject`
            connect_timeout_ms: time to wait for successful connection (ms)
                before considering it a failure
            read_timeout_ms: time to wait for each read (primary effect is to
                determine the application's responsivity when asked to finish;
                see :class:`WhiskerMainSocketListener`)
            name: (name for :class:`StatusMixin`)
            sysevent_prefix: default system event prefix (for
                :class:`WhiskerController`)
            kwargs: parameters to superclass
        """
        super().__init__(parent=parent, name=name, logger=log, **kwargs)
        self.state = ThreadOwnerState.STOPPED
        self.is_connected = False

        self.mainsockthread = QThread(self)
        self.mainsock = WhiskerMainSocketListener(
            server,
            main_port,
            connect_timeout_ms=connect_timeout_ms,
            read_timeout_ms=read_timeout_ms,
            parent=None)  # must be None as it'll go to a different thread
        self.mainsock.moveToThread(self.mainsockthread)

        self.taskthread = QThread(self)
        self.controller = WhiskerController(server,
                                            sysevent_prefix=sysevent_prefix)
        self.controller.moveToThread(self.taskthread)
        self.task = task
        # debug_object(self)
        # debug_thread(self.taskthread)
        # debug_object(self.controller)
        # debug_object(task)
        self.task.moveToThread(self.taskthread)
        # debug_object(self.controller)
        # debug_object(task)
        self.task.set_controller(self.controller)

        # Connect object and thread start/stop events
        # ... start sequence
        self.taskthread.started.connect(self.task.thread_started)
        self.mainsockthread.started.connect(self.mainsock.start)
        # ... stop
        self.mainsock_finish_requested.connect(self.mainsock.stop,
                                               type=Qt.DirectConnection)  # NB!
        self.mainsock.finished.connect(self.mainsockthread.quit)
        self.mainsockthread.finished.connect(self._mainsockthread_finished)
        self.controller_finish_requested.connect(self.task.stop)
        self.task.finished.connect(self.controller.task_finished)
        self.controller.finished.connect(self.taskthread.quit)
        self.taskthread.finished.connect(self._taskthread_finished)

        # Status
        self.mainsock.error_sent.connect(self.error_sent)
        self.mainsock.status_sent.connect(self.status_sent)
        self.controller.error_sent.connect(self.error)
        self.controller.status_sent.connect(self.status_sent)
        self.task.error_sent.connect(self.error)
        self.task.status_sent.connect(self.status_sent)

        # Network communication
        self.mainsock.line_received.connect(self.controller.main_received)
        self.controller.connected.connect(self._on_connect)
        self.controller.connected.connect(self.task.on_connect)
        self.controller.message_received.connect(self.message_received)  # different thread  # noqa
        self.controller.event_received.connect(self.event_received)  # different thread  # noqa
        self.controller.event_received.connect(self.task.on_event)  # same thread  # noqa
        self.controller.key_event_received.connect(self.task.on_key_event)  # same thread  # noqa
        self.controller.client_message_received.connect(self.task.on_client_message)  # same thread  # noqa
        self.controller.pingack_received.connect(self.pingack_received)  # different thread  # noqa
        self.controller.warning_received.connect(self.task.on_warning)  # same thread  # noqa
        self.controller.error_received.connect(self.task.on_error)  # same thread  # noqa
        self.controller.syntax_error_received.connect(
            self.task.on_syntax_error)  # same thread  # noqa

        # Abort events
        self.mainsock.disconnected.connect(self._on_disconnect)
        self.controller.disconnected.connect(self._on_disconnect)

        # Other
        self.ping_requested.connect(self.controller.ping)

    # -------------------------------------------------------------------------
    # General state control
    # -------------------------------------------------------------------------

    def is_running(self) -> bool:
        """
        Are any of our worker threads starting/running/stopping?
        """
        running = self.state != ThreadOwnerState.STOPPED
        self.debug("is_running: {} (state: {})".format(running,
                                                       self.state.name))
        return running

    def _set_state(self, state: ThreadOwnerState) -> None:
        """
        Internal function to set the thread state flag.
        """
        self.debug("state: {} -> {}".format(self.state, state))
        self.state = state

    def report_status(self) -> None:
        """
        Print current thread/server state to the log.
        """
        self.status("state: {}".format(self.state))
        self.status("connected to server: {}".format(self.is_connected))

    # -------------------------------------------------------------------------
    # Starting
    # -------------------------------------------------------------------------

    def start(self) -> None:
        """
        Start our worker threads.
        """
        self.debug("WhiskerOwner: start")
        if self.state != ThreadOwnerState.STOPPED:
            self.error("Can't start: state is: {}".format(self.state.name))
            return
        self.taskthread.start()
        self.mainsockthread.start()
        self._set_state(ThreadOwnerState.RUNNING)

    # -------------------------------------------------------------------------
    # Stopping
    # -------------------------------------------------------------------------

    # noinspection PyArgumentList
    @pyqtSlot()
    @exit_on_exception
    def _on_disconnect(self) -> None:
        """
        Slot called when the Whisker server disconnects.
        """
        self.debug("WhiskerOwner: on_disconnect")
        self.is_connected = False
        self.disconnected.emit()
        if self.state == ThreadOwnerState.STOPPING:
            return
        self.stop()

    def stop(self) -> None:
        """
        Called by the GUI when we want to stop.
        """
        self.info("Stop requested [previous state: {}]".format(
            self.state.name))
        if self.state == ThreadOwnerState.STOPPED:
            self.error("Can't stop: was already stopped")
            return
        self._set_state(ThreadOwnerState.STOPPING)
        self.controller_finish_requested.emit()  # -> self.task.stop
        self.mainsock_finish_requested.emit()  # -> self.mainsock.stop

    # noinspection PyArgumentList
    @pyqtSlot()
    @exit_on_exception
    def _mainsockthread_finished(self) -> None:
        """
        Slot called when our main-socket thread has finished.
        """
        self.debug("stop: main socket thread stopped")
        self._check_everything_finished()

    # noinspection PyArgumentList
    @pyqtSlot()
    @exit_on_exception
    def _taskthread_finished(self) -> None:
        """
        Slot called when our task thread has finished.
        """
        self.debug("stop: task thread stopped")
        self._check_everything_finished()

    def _check_everything_finished(self) -> None:
        """
        Have all out threads finished? If so, emit the ``finished`` signal.
        """
        if self.mainsockthread.isRunning() or self.taskthread.isRunning():
            return
        self._set_state(ThreadOwnerState.STOPPED)
        self.finished.emit()

    # -------------------------------------------------------------------------
    # Other
    # -------------------------------------------------------------------------

    # noinspection PyArgumentList
    @pyqtSlot()
    @exit_on_exception
    def _on_connect(self) -> None:
        """
        Slot called when we are fully connected to the Whisker server.
        """
        self.status("Fully connected to Whisker server")
        self.is_connected = True
        self.connected.emit()

    def ping(self) -> None:
        """
        Pings the Whisker server.
        """
        if not self.is_connected:
            self.warning("Won't ping: not connected")
            return
        self.ping_requested.emit()


# =============================================================================
# Main socket listener
# =============================================================================

class WhiskerMainSocketListener(QObject, StatusMixin):
    """
    Whisker thread (A) (see :class:`WhiskerOwner`).
    Listens to the main socket.
    
    Signals:

    - ``finished()``
    - ``disconnected()``
    - ``line_received(msg: str, timestamp: arrow.Arrow)``
    """
    finished = pyqtSignal()
    disconnected = pyqtSignal()
    line_received = pyqtSignal(str, arrow.Arrow)

    def __init__(self,
                 server: str,
                 port: int,
                 parent: QObject = None,
                 connect_timeout_ms: int = 5000,
                 read_timeout_ms: int = 100,
                 name: str = "whisker_mainsocket",
                 **kwargs) -> None:
        """
        
        Args:
            server: host name or IP address of Whisker server
            port: main port number for Whisker server
            parent: optional parent :`QObject`
            connect_timeout_ms: time to wait for successful connection (ms)
                before considering it a failure
            read_timeout_ms: time to wait for each read (primary effect is to
                determine the application's responsivity when asked to finish;
                see :class:`WhiskerMainSocketListener`)
            name: (name for :class:`StatusMixin`)
            kwargs: parameters to superclass 
        """
        super().__init__(parent=parent, name=name, logger=log, **kwargs)
        self.server = server
        self.port = port
        self.connect_timeout_ms = connect_timeout_ms
        self.read_timeout_ms = read_timeout_ms

        self.finish_requested = False
        self.residual = ''
        self.socket = None
        self.running = False
        # Don't create the socket immediately; we're going to be moved to
        # another thread.

    # noinspection PyArgumentList
    @pyqtSlot()
    def start(self) -> None:
        """
        - Connect to the Whisker server via the main socket.
        - Enter a blocking loop waiting for input and periodically checking
          for a "please finish" signal (with a frequency determined by
          :attr:`read_timeout_ms`.
        - Eventually call :func:`WhiskerMainSocketListener.finish`.
        """
        # Must be separate from __init__, or signals won't be connected yet.
        self.finish_requested = False
        self.status("Connecting to {}:{} with timeout {} ms".format(
            self.server, self.port, self.connect_timeout_ms))
        self.socket = QTcpSocket(self)
        # noinspection PyUnresolvedReferences
        self.socket.disconnected.connect(self.disconnected)
        self.socket.connectToHost(self.server, self.port)
        if not self.socket.waitForConnected(self.connect_timeout_ms):
            errmsg = "Socket error {}".format(get_socket_error(self.socket))
            self.error(errmsg)
            self._finish()
            return
        self.info("Connected to {}:{}".format(self.server, self.port))
        disable_nagle(self.socket)
        # Main blocking loop
        self.running = True
        while not self.finish_requested:
            # self.debug("ping")
            if self.socket.waitForReadyRead(self.read_timeout_ms):
                # data is now ready
                data = self.socket.readAll()  # type: QByteArray
                # log.critical(repr(data))
                # log.critical(repr(type(data)))  # <class 'PyQt5.QtCore.QByteArray'> under PyQt5  # noqa

                # for PySide:
                # - readAll() returns a QByteArray; bytes() fails; str() is OK
                # strdata = str(data)

                # for PyQt5:
                # - readAll() returns a QByteArray again;
                # - however, str(data) looks like "b'Info: ...\\n'"
                strdata = data.data().decode(ENCODING)  # this works

                # log.critical(repr(strdata))
                self._process_data(strdata)
        self.running = False
        self.info("WhiskerMainSocketListener: main event loop complete")
        self._finish()

    # noinspection PyArgumentList
    @pyqtSlot()
    @exit_on_exception
    def stop(self) -> None:
        """
        Stop (by requesting a finish and waiting for the blocking loop in
        :func:`WhiskerMainSocketListener.start` to notice.
        """
        self.debug("WhiskerMainSocketListener: stop")
        if not self.running:
            self.error(
                "WhiskerMainSocketListener: stop requested, but not running")
        self.finish_requested = True

    def _sendline_mainsock(self, msg: str) -> None:
        """
        Send an EOL-terminated message to the server via the main socket.
        """
        if not is_socket_connected(self.socket):
            self.error("Can't send through a closed socket")
            return
        self.debug("Sending to server (MAIN): {}".format(msg))
        final_str = msg + EOL
        data_bytes = final_str.encode(ENCODING)
        self.socket.write(data_bytes)
        self.socket.flush()

    def _finish(self) -> None:
        """
        Clean up and emit the ``finished`` signal.
        """
        if is_socket_connected(self.socket):
            self.socket.close()
        self.info("WhiskerMainSocketListener: finished")
        self.finished.emit()

    def _process_data(self, data: str) -> None:
        """
        Adds the incoming data to any stored residual, splits it into lines,
        and sends each line out via the ``line_received`` signal (on to the
        receiver).
        """
        self.debug("incoming: {}".format(repr(data)))
        timestamp = arrow.now()
        data = self.residual + data
        fragments = data.split(EOL)
        lines = fragments[:-1]
        self.residual = fragments[-1]
        for line in lines:
            self.debug("incoming line: {}".format(line))
            if line == PING:
                self._sendline_mainsock(PING_ACK)
                self.status("Ping received from server")
                return
            self.line_received.emit(line, timestamp)


# =============================================================================
# Object to talk to task and to immediate socket
# =============================================================================

class WhiskerController(QObject, StatusMixin, WhiskerApi):
    """
    Controls the Whisker immediate socket.

    - Encapsulates the Whisker API.
    - Lives in Whisker thread (B) (see :class:`WhiskerOwner`).
    - Emits signals that can be processed by the Whisker task (derived from
      :class:`WhiskerQtTask`).
    
    Signals:

    - ``finished()``
    - ``connected()``
    - ``disconnected()``
    - ``message_received(str, arrow.Arrow, int)``
    - ``event_received(str, arrow.Arrow, int)``
    - ``key_event_received(str, arrow.Arrow, int)``
    - ``client_message_received(int, str, arrow.Arrow, int)``
    - ``warning_received(str, arrow.Arrow, int)``
    - ``syntax_error_received(str, arrow.Arrow, int)``
    - ``error_received(str, arrow.Arrow, int)``
    - ``pingack_received(arrow.Arrow, int)``

    """
    finished = pyqtSignal()
    connected = pyqtSignal()
    disconnected = pyqtSignal()
    message_received = pyqtSignal(str, arrow.Arrow, int)
    event_received = pyqtSignal(str, arrow.Arrow, int)
    key_event_received = pyqtSignal(str, arrow.Arrow, int)
    client_message_received = pyqtSignal(int, str, arrow.Arrow, int)
    warning_received = pyqtSignal(str, arrow.Arrow, int)
    syntax_error_received = pyqtSignal(str, arrow.Arrow, int)
    error_received = pyqtSignal(str, arrow.Arrow, int)
    pingack_received = pyqtSignal(arrow.Arrow, int)

    def __init__(self,
                 server: str,
                 parent: QObject = None,
                 connect_timeout_ms: int = 5000,
                 read_timeout_ms: int = 500,
                 name: str = "whisker_controller",
                 sysevent_prefix: str = "sys_",
                 **kwargs) -> None:
        """
        Args:
            server: host name or IP address of Whisker server
            parent: optional parent :`QObject`
            connect_timeout_ms: time to wait for successful connection (ms)
                before considering it a failure
            read_timeout_ms: time to wait for each read (primary effect is to
                determine the application's responsivity when asked to finish;
                see :class:`WhiskerMainSocketListener`)
            name: (name for :class:`StatusMixin`)
            sysevent_prefix: default system event prefix (for
                :class:`WhiskerController`)
            kwargs: parameters to superclass
        """
        super().__init__(
            # For QObject:
            parent=parent,

            # For StatusMixin:
            name=name,
            logger=log,

            # For WhiskerApi:
            whisker_immsend_get_reply_fn=self._get_immsock_response,
            sysevent_prefix=sysevent_prefix,

            # Anyone else?
            **kwargs
        )
        self.server = server
        self.connect_timeout_ms = connect_timeout_ms
        self.read_timeout_ms = read_timeout_ms

        self.immport = None  # type: Optional[int]
        self.code = None  # type: Optional[str]
        self.immsocket = None  # type: Optional[QTcpSocket]
        self.residual = ''

    # noinspection PyArgumentList
    @pyqtSlot(str, arrow.Arrow)
    @exit_on_exception
    def main_received(self, msg: str, timestamp: arrow.Arrow) -> None:
        """
        Slot to process a message that has come in from the main socket.

        Args:
            msg: raw message
            timestamp: server timestamp
        """
        gre = CompiledRegexMemory()
        # self.debug("main_received: {}".format(msg))

        # 0. Ping has already been dealt with.
        # 1. Deal with immediate socket connection internally.
        if gre.search(IMMPORT_REGEX, msg):
            self.immport = int(gre.group(1))
            return

        if gre.search(CODE_REGEX, msg):
            code = gre.group(1)
            self.immsocket = QTcpSocket(self)
            # noinspection PyUnresolvedReferences
            self.immsocket.disconnected.connect(self.disconnected)
            self.debug(
                "Connecting immediate socket to {}:{} with timeout {}".format(
                    self.server, self.immport, self.connect_timeout_ms))
            self.immsocket.connectToHost(self.server, self.immport)
            if not self.immsocket.waitForConnected(self.connect_timeout_ms):
                errmsg = "Immediate socket error {}".format(
                    get_socket_error(self.immsocket))
                self.error(errmsg)
                self.finish()
            self.debug("Connected immediate socket to "
                       "{}:{}".format(self.server, self.immport))
            disable_nagle(self.immsocket)
            self.command("Link {}".format(code))
            self.connected.emit()
            return

        # 2. Get timestamp.
        (msg, whisker_timestamp) = split_timestamp(msg)

        # 3. Send the message to a general-purpose receiver
        self.message_received.emit(msg, timestamp, whisker_timestamp)

        # 4. Send the message to specific-purpose receivers.
        if gre.search(EVENT_REGEX, msg):
            event = gre.group(1)
            if self.process_backend_event(event):
                return
            self.event_received.emit(event, timestamp, whisker_timestamp)
        elif gre.search(KEY_EVENT_REGEX, msg):
            key = gre.group(1)
            self.key_event_received.emit(key, timestamp, whisker_timestamp)
        elif gre.search(CLIENT_MESSAGE_REGEX, msg):
            source_client_num = int(gre.group(1))
            client_msg = gre.group(2)
            self.client_message_received.emit(source_client_num, client_msg,
                                              timestamp, whisker_timestamp)
        elif WARNING_REGEX.match(msg):
            self.warning_received.emit(msg, timestamp, whisker_timestamp)
        elif SYNTAX_ERROR_REGEX.match(msg):
            self.syntax_error_received.emit(msg, timestamp, whisker_timestamp)
        elif ERROR_REGEX.match(msg):
            self.error_received.emit(msg, timestamp, whisker_timestamp)
        elif msg == PING_ACK:
            self.pingack_received.emit(timestamp, whisker_timestamp)

    # noinspection PyArgumentList
    @pyqtSlot()
    @exit_on_exception
    def task_finished(self) -> None:
        """
        Slot called when the task says that it's finished.
        """
        self.debug("Task reports that it is finished")
        self._close_immsocket()
        self.finished.emit()

    def _sendline_immsock(self, *args) -> None:
        msg = msg_from_args(*args)
        self.debug("Sending to server (IMM): {}".format(msg))
        final_str = msg + EOL
        data_bytes = final_str.encode(ENCODING)
        self.immsocket.write(data_bytes)
        self.immsocket.waitForBytesWritten(INFINITE_WAIT)
        # http://doc.qt.io/qt-4.8/qabstractsocket.html
        self.immsocket.flush()

    def _getline_immsock(self) -> str:
        """
        Get one line from the socket, and returns it. Blocking.

        We must also respond to the possibility that the socket has been
        forcibly closed.
        """
        data = self.residual
        while EOL not in data and is_socket_connected(self.immsocket):
            # get more data from socket
            # self.debug("WAITING FOR DATA")
            self.immsocket.waitForReadyRead(INFINITE_WAIT)
            # self.debug("DATA READY. READING IT.")
            newdata_bytearray = self.immsocket.readAll()  # type: QByteArray
            newdata_str = newdata_bytearray.data().decode(ENCODING)
            data += newdata_str
            # self.debug("DATA: {}".format(repr(data)))
        # self.debug("DATA COMPLETE")
        if EOL in data:
            eol_index = data.index(EOL)
            line = data[:eol_index]
            self.residual = data[eol_index + EOL_LEN:]
        else:
            line = ''  # socket is closed
            self.residual = data  # probably blank!
        self.debug("Reply from server (IMM): {}".format(line))
        return line

    def _get_immsock_response(self, *args) -> Optional[str]:
        """
        Builds its arguments into a string, sends it as a command to the
        Whisker server via the immediate socket, waits for a reply, and
        returns it. (Returns ``None`` if the socket isn't connected.)
        """
        if not self.is_connected():
            self.error("Not connected")
            return None
        self._sendline_immsock(*args)
        reply = self._getline_immsock()
        return reply

    def is_connected(self) -> bool:
        """
        Is our immediate socket connected?

        (If the immediate socket is running, the main socket should be too.)
        """
        return is_socket_connected(self.immsocket)

    def _close_immsocket(self) -> None:
        """
        Close the immediate socket.
        """
        if is_socket_connected(self.immsocket):
            self.immsocket.close()

    def ping(self) -> None:
        """
        Ping the server.

        Override :func:`WhiskerApi.ping` so we can emit a signal on success.
        """
        # override WhiskerApi.ping() so we can emit a signal on success
        reply, whisker_timestamp = self._immresp_with_timestamp(PING)
        if reply == PING_ACK:
            timestamp = arrow.now()
            self.pingack_received.emit(timestamp, whisker_timestamp)


# =============================================================================
# Object from which Whisker tasks should be subclassed
# =============================================================================

class WhiskerQtTask(QObject, StatusMixin):
    """
    Base class for building a Whisker task itself. You should derive from
    this.

    - Lives in Whisker thread (B) (see :class:`WhiskerOwner`).

    Signals:

    - ``finished()``

    """
    finished = pyqtSignal()  # emit from stop() function when all done

    def __init__(self, parent: QObject = None,
                 name: str = "whisker_task", **kwargs) -> None:
        """
        Args:
            parent: optional parent :`QObject`
            name: (name for :class:`StatusMixin`)
            kwargs: parameters to superclass
        """
        super().__init__(name=name, logger=log, parent=parent, **kwargs)
        self.whisker = None

    def set_controller(self, controller: WhiskerController) -> None:
        """
        Called by :class:`WhiskerOwner`. No need to override.
        """
        self.whisker = controller

    # noinspection PyMethodMayBeStatic,PyArgumentList
    @pyqtSlot()
    def thread_started(self) -> None:
        """
        Slot called from the :func:`WhiskerOwner.taskthread.started` signal,
        which is called indirectly by :func:`WhiskerOwner.start`.

        - Use this for additional setup if required.
        - No need to override in simple situations.
        """
        pass

    # noinspection PyArgumentList
    @pyqtSlot()
    def stop(self) -> None:
        """
        Called by the :class:`WhiskerOwner` when we should stop.

        - When we've done what we need to, we should emit the ``finished``
          signal.
        - No need to override in simple situations.

        NOTE: if you think this function is not being called, the likely reason
        is not a Qt signal/slot failure, but that the thread is BUSY, e.g.
        with its immediate socket. (It'll be busy via the derived class, not
        via the code here, which does no waiting!)
        """
        self.info("WhiskerQtTask: stopping")
        self.finished.emit()

    @exit_on_exception
    def on_connect(self) -> None:
        """
        The :class:`WhiskerOwner` makes this slot get called when the
        :class:`WhiskerController` is connected.

        You should override this.
        """
        self.warning("on_connect: YOU SHOULD OVERRIDE THIS")

    # noinspection PyUnusedLocal,PyArgumentList
    @pyqtSlot(str, arrow.Arrow, int)
    @exit_on_exception
    def on_event(self, event: str, timestamp: arrow.Arrow,
                 whisker_timestamp_ms: int) -> None:
        """
        General Whisker events come here.

        Args:
            event: incoming event
            timestamp: time of receipt by client
            whisker_timestamp_ms: server's timestamp (ms)
        """
        # You should override this
        msg = "SHOULD BE OVERRIDDEN. EVENT: {}".format(event)
        self.status(msg)

    # noinspection PyUnusedLocal,PyArgumentList
    @pyqtSlot(str, arrow.Arrow, int)
    @exit_on_exception
    def on_key_event(self, key_event: str, timestamp: arrow.Arrow,
                     whisker_timestamp_ms: int) -> None:
        """
        Keyboard events come here.

        Args:
            key_event: incoming event
            timestamp: time of receipt by client
            whisker_timestamp_ms: server's timestamp (ms)
        """
        msg = "SHOULD BE OVERRIDDEN. KEY EVENT: {}".format(key_event)
        self.status(msg)

    # noinspection PyUnusedLocal,PyArgumentList
    @pyqtSlot(int, str, arrow.Arrow, int)
    @exit_on_exception
    def on_client_message(self,
                          source_client_num: int,
                          client_msg: str,
                          timestamp: arrow.Arrow,
                          whisker_timestamp_ms: int) -> None:
        """
        Client messages (from other clients) come here.

        Args:
            source_client_num: client number of sender
            client_msg: message
            timestamp: time of receipt by client
            whisker_timestamp_ms: server's timestamp (ms)
        """
        msg = "SHOULD BE OVERRIDDEN. CLIENT MESSAGE FROM CLIENT {}: {}".format(
            source_client_num, repr(client_msg))
        self.status(msg)

    # noinspection PyUnusedLocal,PyArgumentList
    @pyqtSlot(str, arrow.Arrow, int)
    @exit_on_exception
    def on_warning(self, msg: str, timestamp: arrow.Arrow,
                   whisker_timestamp_ms: int) -> None:
        """
        Warnings from the server come here.

        Args:
            msg: incoming event
            timestamp: time of receipt by client
            whisker_timestamp_ms: server's timestamp (ms)
        """
        self.warning(msg)

    # noinspection PyUnusedLocal,PyArgumentList
    @pyqtSlot(str, arrow.Arrow, int)
    @exit_on_exception
    def on_error(self, msg: str, timestamp: arrow.Arrow,
                 whisker_timestamp_ms: int) -> None:
        """
        Error reports from the server come here.

        Args:
            msg: incoming event
            timestamp: time of receipt by client
            whisker_timestamp_ms: server's timestamp (ms)
        """
        self.error(msg)

    # noinspection PyUnusedLocal,PyArgumentList
    @pyqtSlot(str, arrow.Arrow, int)
    @exit_on_exception
    def on_syntax_error(self, msg: str, timestamp: arrow.Arrow,
                        whisker_timestamp_ms: int) -> None:
        """
        Syntax error reports from the server come here.

        Args:
            msg: incoming event
            timestamp: time of receipt by client
            whisker_timestamp_ms: server's timestamp (ms)
        """
        self.error(msg)



#!/usr/bin/env python

"""
whisker/twistedclient.py

===============================================================================

    Copyright © 2011-2020 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of the Whisker Python client library.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Event-driven framework for Whisker Python clients using Twisted.**

- Created: 18 Aug 2011
- Last update: 10 Feb 2016
"""

import logging
import re
import socket
from typing import Generator, Optional, Union

from twisted.internet import reactor
# from twisted.internet.stdio import StandardIO
from twisted.internet.protocol import ClientFactory
from twisted.internet.tcp import Connector  # for type hints
from twisted.protocols.basic import LineReceiver

from whisker.api import (
    CLIENT_MESSAGE_PREFIX,
    ERROR_PREFIX,
    EVENT_PREFIX,
    INFO_PREFIX,
    KEY_EVENT_PREFIX,
    msg_from_args,
    on_off_to_boolean,
    split_timestamp,
    SYNTAX_ERROR_PREFIX,
    WARNING_PREFIX,
    WhiskerApi,
)
from whisker.socket import (
    get_port,
    socket_receive,
    socket_sendall,
)

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


# =============================================================================
# Event-driven Whisker task class. Use this one.
# =============================================================================

class WhiskerTwistedTask(object):
    """
    Base class for Whisker clients using the Twisted socket system.

    - Contains ``self.whisker``, an instance of
      :class:`whisker.api.WhiskerApi`.
    - Specimen usage: see ``test_twisted.py``.
    """

    def __init__(self) -> None:
        self.server = None
        self.mainport = None
        self.immport = None
        self.code = None
        self.mainsocket = None
        self.immsocket = None
        self.mainfactory = WhiskerMainPortFactory(self)
        self.whisker = WhiskerApi(
            whisker_immsend_get_reply_fn=self.send_and_get_reply)

    @classmethod
    def set_verbose_logging(cls, verbose: bool) -> None:
        """
        Sets the Python log level for this module.

        Args:
            verbose: be verbose?
        """
        if verbose:
            log.setLevel(logging.DEBUG)
        else:
            log.setLevel(logging.INFO)

    def connect(self, server: str, port: Union[str, int]) -> None:
        """
        Connects to the Whisker server.

        Args:
            server: Whisker server hostname/IP address
            port: Whisker main TCP/IP port number
        """
        self.server = server
        self.mainport = get_port(port)
        log.info(
            "Attempting to connect to Whisker server {s} on port {p}".format(
                s=self.server,
                p=self.mainport
            ))
        # noinspection PyUnresolvedReferences
        reactor.connectTCP(self.server, self.mainport, self.mainfactory)

    def _connect_immediate(self) -> None:
        """
        Connect the Whisker immediate socket.
        """
        # Twisted really hates blocking.
        # So we need to do some special things here.
        self.immsocket = WhiskerImmSocket(self)
        log.info(
            "Attempting to connect to Whisker server {s} on immediate "
            "port {p}".format(
                s=self.server,
                p=self.immport
            ))
        self.immsocket.connect(self.server, self.immport)
        if not self.immsocket.connected:
            log.error("ERROR creating/connecting immediate socket: " +
                      str(self.immsocket.error))
        log.info("Connected to immediate port " + str(self.immport) +
                 " on server " + self.server)
        self.immsocket.send_and_get_reply("Link " + self.code)
        log.info("Server fully connected.")
        self.fully_connected()

        # sleeptime = 0.1
        # log.info("Sleeping for " + str(sleeptime) +
        #                  " seconds as the Nagle-disabling feature of Python "
        #                  "isn't working properly...")
        # time.sleep(sleeptime)
        # The Nagle business isn't working; the Link packet is getting
        # amalgamated with anything the main calling program starts to send.
        # So pause.

    def fully_connected(self) -> None:
        """
        The Whisker server is now fully connected.

        Override this, e.g. to start the task.
        """
        pass

    def send(self, *args) -> None:
        """
        Builds its arguments into a string and sends that as a command to the
        Whisker server via the main socket.
        """
        if not self.mainsocket:
            log.error("can't send without a mainsocket")
            return
        msg = msg_from_args(*args)
        self.mainsocket.send(msg)

    def send_and_get_reply(self, *args) -> Optional[str]:
        """
        Builds its arguments into a string, sends that as a command to the
        Whisker server via the immediate socket, blocks and waits for the
        reply, and returns that reply. (Returns ``None`` if not connected.)
        """
        if not self.immsocket:
            log.error("can't send_and_get_reply without an immsocket")
            return
        reply = self.immsocket.send_and_get_reply(*args)
        return reply

    def incoming_message(self, msg: str) -> None:
        """
        Processes an incoming message from the Whisker server (via the main
        socket).
        """
        # log.debug("INCOMING MESSAGE: " + str(msg))
        handled = False
        if not self.immport:
            m = re.search(r"^ImmPort: (\d+)", msg)
            if m:
                self.immport = get_port(m.group(1))
                handled = True
        if not self.code:
            m = re.search(r"^Code: (\w+)", msg)
            if m:
                self.code = m.group(1)
                handled = True
        if (not self.immsocket) and (self.immport and self.code):
            self._connect_immediate()
        if handled:
            return

        (msg, timestamp) = split_timestamp(msg)

        if msg == "Ping":
            # If the server has sent us a Ping, acknowledge it.
            self.send("PingAcknowledged")
            return

        if msg.startswith(EVENT_PREFIX):
            # The server has sent us an event.
            event = msg[len(EVENT_PREFIX):]
            self.incoming_event(event, timestamp)
            return

        if msg.startswith(KEY_EVENT_PREFIX):
            kmsg = msg[len(KEY_EVENT_PREFIX):]
            # key on|off document
            m = re.match(r"(\w+)\s+(\w+)\s+(\w+)", kmsg)
            if m:
                key = m.group(1)
                depressed = on_off_to_boolean(m.group(2))
                document = m.group(3)
                self.incoming_key_event(key, depressed, document, timestamp)
            return

        if msg.startswith(CLIENT_MESSAGE_PREFIX):
            cmsg = msg[len(CLIENT_MESSAGE_PREFIX):]
            # fromclientnum message
            m = re.match(r"(\w+)\s+(.+)", cmsg)
            if m:
                try:
                    fromclientnum = int(m.group(1))
                    clientmsg = m.group(2)
                    self.incoming_client_message(fromclientnum, clientmsg,
                                                 timestamp)
                except (TypeError, ValueError):
                    pass
            return

        if msg.startswith(INFO_PREFIX):
            self.incoming_info(msg)
            return

        if msg.startswith(WARNING_PREFIX):
            self.incoming_warning(msg)
            return

        if msg.startswith(SYNTAX_ERROR_PREFIX):
            self.incoming_syntax_error(msg)
            return

        if msg.startswith(ERROR_PREFIX):
            self.incoming_error(msg)
            return

        log.debug("Unhandled incoming message: " + str(msg))

    def incoming_event(self, event: str, timestamp: int = None) -> None:
        """
        General Whisker event received.

        Override this function.

        Args:
            event: event
            timestamp: server timestamp (ms)
        """
        log.debug("UNHANDLED EVENT: {e} (timestamp={t}".format(
            e=event,
            t=timestamp
        ))

    # noinspection PyMethodMayBeStatic
    def incoming_client_message(self, fromclientnum: int, msg: str,
                                timestamp: int = None) -> None:
        """
        Client message (from another client) received.

        Override this function.

        Args:
            fromclientnum: source client number
            msg: message
            timestamp: server timestamp (ms)
        """
        log.debug(
            "UNHANDLED CLIENT MESSAGE from client {c}: {m} "
            "(timestamp={t})".format(
                c=fromclientnum,
                m=msg,
                t=timestamp
            ))

    # noinspection PyMethodMayBeStatic
    def incoming_key_event(self, key: str, depressed: bool, document: str,
                           timestamp: int = None) -> None:
        """
        Keyboard event received.

        Override this function.

        Args:
            key: which key?
            depressed: was it depressed (or released)?
            document: source display document
            timestamp: server timestamp (ms)
        """
        log.debug(
            "UNHANDLED KEY EVENT: key {k} {dr} (document={d}, "
            "timestamp={t})".format(
                k=key,
                dr="depressed" if depressed else "released",
                d=document,
                t=timestamp
            ))

    # noinspection PyMethodMayBeStatic
    def incoming_info(self, msg: str) -> None:
        """
        Information message received from Whisker server.

        Args:
            msg: message
        """
        log.info(msg)

    # noinspection PyMethodMayBeStatic
    def incoming_warning(self, msg: str) -> None:
        """
        Warning received from Whisker server.

        Args:
            msg: message
        """
        log.warning(msg)

    # noinspection PyMethodMayBeStatic
    def incoming_error(self, msg: str) -> None:
        """
        Error report received from Whisker server.

        Args:
            msg: message
        """
        log.error(msg)

    # noinspection PyMethodMayBeStatic
    def incoming_syntax_error(self, msg: str) -> None:
        """
        Syntax error report received from Whisker server.

        Args:
            msg: message
        """
        log.error(msg)


class WhiskerMainPortFactory(ClientFactory):
    """
    A Protocol factory for the Whisker main port.
    """

    def __init__(self, task: WhiskerTwistedTask) -> None:
        """
        Args:
            task: instance of :class:`WhiskerTwistedTask`
        """
        self.task = task

    def clientConnectionLost(self, connector: Connector, reason: str) -> None:
        """
        If we get disconnected, reconnect to server.
        """
        log.warning("WhiskerMainPortFactory: disconnected")
        connector.connect()

    def clientConnectionFailed(self, connector: Connector,
                               reason: str) -> None:
        """
        Client connection failed. Stop the reactor.
        """
        log.error("connection failed: " + str(reason))
        reactor.stop()

    def buildProtocol(self, addr: str) -> Optional['WhiskerMainPortProtocol']:
        """
        Build and return the protocol.
        """
        log.debug("WhiskerMainPortFactory: buildProtocol({})".format(addr))
        if self.task.mainsocket:
            log.error("mainsocket already connected")
            return None
        p = WhiskerMainPortProtocol(self.task)
        return p


class WhiskerMainPortProtocol(LineReceiver):
    """
    Line-based Twisted protocol for the Whisker main port.
    """

    delimiter = b"\n"  # MUST BE BYTES, NOT STR!
    # Otherwise, you get a crash ('str' does not support the buffer interface)
    # from within twisted/protocols/basic.py, line 559, when it tries to do
    #   something = bytes_buffer.split(string_delimiter, 1)

    def __init__(self, task: WhiskerTwistedTask,
                 encoding: str = 'ascii') -> None:
        """
        Args:
            task: instance of :class:`WhiskerTwistedTask`
            encoding: encoding to use; normally ``"ascii"``
        """
        self.task = task
        self.task.mainsocket = self
        self.encoding = encoding

    def connectionMade(self) -> None:
        """
        Called when the main port is connected.
        """
        peer = self.transport.getPeer()
        if hasattr(peer, "host") and hasattr(peer, "port"):
            log.info("Connected to main port {p} on server {h}".format(
                h=peer.host,
                p=peer.port
            ))
        else:
            log.debug("Connected to main port")
        self.transport.setTcpNoDelay(True)  # disable Nagle algorithm
        log.debug("Main port: Nagle algorithm disabled (TCP_NODELAY set)")

    def lineReceived(self, data: bytes) -> None:
        """
        Called when data is received on the main port.
        Sends it to :func:`WhiskerTwistedTask.incoming_message` via
        ``self.task``.

        Args:
            data: bytes
        """
        str_data = data.decode(self.encoding)
        log.debug("Main port received: {}".format(str_data))
        self.task.incoming_message(str_data)

    def send(self, data: str) -> None:
        """
        Encodes and sends data to the main port.
        """
        log.debug("Main port sending: {}".format(data))
        self.sendLine(data.encode(self.encoding))

    def rawDataReceived(self, data: bytes) -> None:
        """
        Raw data received. Unused; we use :func:`lineReceived` instead.
        """
        pass


class WhiskerImmSocket(object):
    """
    Whisker Twisted immediate socket handler.

    Uses raw sockets.
    """

    def __init__(self, task: WhiskerTwistedTask) -> None:
        """
        Args:
            task: instance of :class:`WhiskerTwistedTask`
        """
        self.task = task
        self.connected = False
        self.error = ""
        self.immsock = None

    def connect(self, server: str, port: int) -> None:
        """
        Connects the Whisker immediate socket.

        Args:
            server: server hostname/IP address
            port: immediate port number
        """
        log.debug("WhiskerImmSocket: connect")
        proto = socket.getprotobyname("tcp")
        try:
            self.immsock = socket.socket(socket.AF_INET, socket.SOCK_STREAM,
                                         proto)
            self.immsock.connect((server, port))
            self.connected = True
        except socket.error as x:
            self.immsock.close()
            self.immsock = None
            self.error = str(x)
            return

        # Disable the Nagle algorithm:
        self.immsock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        log.debug("Immediate port: Nagle algorithm disabled (TCP_NODELAY set)")
        # Set blocking
        self.immsock.setblocking(True)
        log.debug("Immediate port: set to blocking mode")

    def getlines_immsock(self) -> Generator[str, None, None]:
        """
        Generates lines from the immediate socket.

        Yields:
            lines from the socket
        """
        # log.debug("WhiskerImmSocket: getlines_immsock")
        # http://stackoverflow.com/questions/822001/python-sockets-buffering
        buf = socket_receive(self.immsock)
        done = False
        while not done:
            if "\n" in buf:
                (line, buf) = buf.split("\n", 1)
                yield line
            else:
                more = socket_receive(self.immsock)
                if not more:
                    done = True
                else:
                    buf += more
        if buf:
            yield buf

    def send_and_get_reply(self, *args) -> str:
        """
        Builds its arguments into a string; sends it to Whisker via the
        immediate socket; gets the reply; returns it.
        """
        msg = msg_from_args(*args)
        log.debug("Immediate socket sending: " + msg)
        socket_sendall(self.immsock, msg + "\n")
        reply = next(self.getlines_immsock())
        log.debug("Immediate socket reply: " + reply)
        return reply



#!/usr/bin/env python

"""
whisker/sqlalchemy.py

===============================================================================

    Copyright © 2011-2020 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of the Whisker Python client library.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**SQLAlchemy helper functions for Whisker tasks.**

"""

from contextlib import contextmanager
import logging
from typing import Any, Dict, Generator, Tuple

from sqlalchemy import create_engine, event
from sqlalchemy.engine.base import Engine  # for type hints
from sqlalchemy.orm import scoped_session, Session,  sessionmaker

log = logging.getLogger(__name__)


# =============================================================================
# Functions to get SQLAlchemy database session, etc.
# =============================================================================

def get_database_engine(settings: Dict[str, Any],
                        unbreak_sqlite_transactions: bool = True,
                        pool_pre_ping: bool = True) -> Engine:
    """
    Get an SQLAlchemy database :class:`Engine` from a simple definition.

    Args:
        settings: a dictionary with the following keys:

            - ``url``: a string
            - ``echo``: a boolean
            - ``connect_args``: a dictionary

            All are passed to SQLAlchemy's :func:`create_engine` function.

        unbreak_sqlite_transactions: hook in events to unbreak SQLite
            transaction support? (Detailed in
            ``sqlalchemy/dialects/sqlite/pysqlite.py``; see "Serializable
            isolation / Savepoints / Transactional DDL".)

        pool_pre_ping: boolean; requires SQLAlchemy 1.2

    Returns:
        an SQLAlchemy :class:`Engine`

    """
    database_url = settings['url']
    engine = create_engine(
        database_url,
        echo=settings['echo'],
        connect_args=settings['connect_args'],
        pool_pre_ping=pool_pre_ping  # requires SQLAlchemy 1.2
    )
    sqlite = database_url.startswith("sqlite:")
    if not sqlite or not unbreak_sqlite_transactions:
        return engine

    # Hook in events to unbreak SQLite transaction support
    # Detailed in sqlalchemy/dialects/sqlite/pysqlite.py; see
    # "Serializable isolation / Savepoints / Transactional DDL"

    # noinspection PyUnusedLocal
    @event.listens_for(engine, "connect")
    def do_connect(dbapi_connection, connection_record):
        # disable pysqlite's emitting of the BEGIN statement entirely.
        # also stops it from emitting COMMIT before any DDL.
        dbapi_connection.isolation_level = None

    @event.listens_for(engine, "begin")
    def do_begin(conn):
        # emit our own BEGIN
        conn.execute("BEGIN")

    return engine


# -----------------------------------------------------------------------------
# Plain functions: not thread-aware; generally AVOID these
# -----------------------------------------------------------------------------

# noinspection PyPep8Naming
def get_database_session_thread_unaware(settings: Dict[str, Any]) -> Session:
    """
    Returns an SQLAlchemy database session.

    .. warning:: DEPRECATED: this function is not thread-aware.

    Args:
        settings: passed to :func:`get_database_engine`

    Returns:
        an SQLAlchemy :class:`Session`

    """
    log.warning("get_database_session_thread_unaware() called")
    engine = get_database_engine(settings)
    SessionClass = sessionmaker(bind=engine)
    return SessionClass()


@contextmanager
def session_scope_thread_unaware(
        settings: Dict[str, Any]) -> Generator[Session, None, None]:
    """
    Context manager to provide an SQLAlchemy database session (which
    executes a ``COMMIT`` on success or a ``ROLLBACK`` on failure).

    .. warning:: DEPRECATED: this function is not thread-aware.

    Args:
        settings: passed to :func:`get_database_session_thread_unaware`

    Yields:
        an SQLAlchemy :class:`Session`

    """
    log.warning("session_scope_thread_unaware() called")
    # http://docs.sqlalchemy.org/en/latest/orm/session_basics.html#session-faq-whentocreate  # noqa
    session = get_database_session_thread_unaware(settings)
    # noinspection PyPep8
    try:
        yield session
        session.commit()
    except:
        session.rollback()
        raise
    finally:
        session.close()


# -----------------------------------------------------------------------------
# Thread-scoped versions
# -----------------------------------------------------------------------------
# http://docs.sqlalchemy.org/en/latest/orm/contextual.html
# https://writeonly.wordpress.com/2009/07/16/simple-read-only-sqlalchemy-sessions/  # noqa
# http://docs.sqlalchemy.org/en/latest/orm/session_api.html

# noinspection PyUnusedLocal
def noflush_readonly(*args, **kwargs) -> None:
    """
    Does nothing, and is thereby used to block a database session flush.
    """
    log.debug("Attempt to flush a readonly database session blocked")


# noinspection PyPep8Naming
def get_database_engine_session_thread_scope(
        settings: Dict[str, Any],
        readonly: bool = False,
        autoflush: bool = True) -> Tuple[Engine, Session]:
    """
    Gets a thread-scoped SQLAlchemy :class:`Engine` and :class:`Session`.

    Args:
        settings: passed to :func:`get_database_engine`
        readonly: make the session read-only?
        autoflush: passed to :func:`sessionmaker`

    Returns:
        tuple: ``(engine, session)``

    """
    # The default for a Session is: autoflush=True, autocommit=False
    # http://docs.sqlalchemy.org/en/latest/orm/session_api.html#sqlalchemy.orm.session.Session  # noqa
    if readonly:
        autoflush = False
    engine = get_database_engine(settings)
    session_factory = sessionmaker(bind=engine, autoflush=autoflush)
    SessionClass = scoped_session(session_factory)
    session = SessionClass()
    if readonly:
        session.flush = noflush_readonly
    return engine, session


def get_database_session_thread_scope(*args, **kwargs) -> Session:
    """
    Gets a thread-scoped SQLAlchemy :class:`Session`.

    Args:
        args: positional arguments to
            :func:`get_database_engine_session_thread_scope`
        kwargs: keyword arguments to
            :func:`get_database_engine_session_thread_scope`

    Returns:
        the session

    """
    engine, session = get_database_engine_session_thread_scope(*args, **kwargs)
    return session


@contextmanager
def session_thread_scope(
        settings: Dict[str, Any],
        readonly: bool = False) -> Generator[Session, None, None]:
    """
    Context manager to provide a thread-safe SQLAlchemy database session (which
    executes a ``COMMIT`` on success or a ``ROLLBACK`` on failure).

    Args:
        settings: passed to :func:`get_database_session_thread_scope`
        readonly: passed to :func:`get_database_session_thread_scope`

    Yields:
        an SQLAlchemy :class:`Session`

    """
    session = get_database_session_thread_scope(settings, readonly)
    # noinspection PyPep8
    try:
        yield session
        if not readonly:
            session.commit()
    except:
        if not readonly:
            session.rollback()
        raise
    finally:
        session.close()


# =============================================================================
# Info functions
# =============================================================================

def database_is_sqlite(dbsettings: Dict[str, str]) -> bool:
    """
    Checks the URL in ``dbsettings['url']``: is it an SQLite database?
    """
    database_url = dbsettings['url']
    return database_url.startswith("sqlite:")


def database_is_postgresql(dbsettings: Dict[str, str]) -> bool:
    """
    Checks the URL in ``dbsettings['url']``: is it a PostgreSQL database?
    """
    database_url = dbsettings['url']
    return database_url.startswith("postgresql")
    # ignore colon, since things like "postgresql:", "postgresql+psycopg2:"
    # are all OK


def database_is_mysql(dbsettings: Dict[str, str]) -> bool:
    """
    Checks the URL in ``dbsettings['url']``: is it a MySQL database?
    """
    database_url = dbsettings['url']
    return database_url.startswith("mysql")



#!/usr/bin/env/python

"""
whisker/__init__.py

===============================================================================

    Copyright © 2011-2020 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of the Whisker Python client library.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

The mere existence of this file makes Python treat the directory as a package.

"""

import logging

# Create whisker.__version__:
# noinspection PyPep8Naming
from whisker.version import VERSION as __version__

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())
# http://eric.themoritzfamily.com/learning-python-logging.html
# http://stackoverflow.com/questions/12296214/python-logging-with-a-library-namespaced-packages  # noqa



#!/usr/bin/env python

"""
whisker/test_twisted.py

===============================================================================

    Copyright © 2011-2020 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of the Whisker Python client library.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Command-line tool to test the Whisker Twisted client.**

"""

import argparse
import logging

from twisted.internet import reactor

from cardinal_pythonlib.logs import configure_logger_for_colour
from whisker.api import (
    Pen,
    PenStyle,
    BrushStyle,
    BrushHatchStyle,
    Brush,
    Rectangle,
)
from whisker.constants import DEFAULT_PORT
from whisker.twistedclient import WhiskerTwistedTask

log = logging.getLogger(__name__)

DEFAULT_DISPLAY_NUM = 0
DEFAULT_AUDIO_NUM = 0
DEFAULT_INPUT_LINE = 0
DEFAULT_OUTPUT_LINE = 64
DEFAULT_MEDIA_DIR = r"C:\Program Files\WhiskerControl\Server Test Media"
DEFAULT_BITMAP = "santa_fe.bmp"
DEFAULT_VIDEO = "mediaexample.wmv"
DEFAULT_WAV = "telephone.wav"

AUDIO = "audio"
DISPLAY = "display"
DOC = "doc"
VIDEO = "_video"


class MyWhiskerTwistedTask(WhiskerTwistedTask):
    """
    Class deriving from :class:`whisker.twistedclient.WhiskerTwistedTask` to
    demonstrate the Twisted Whisker client.
    """
    def __init__(self,
                 display_num: int,
                 audio_num: int,
                 input_line: str,
                 output_line: str,
                 media_dir: str,
                 bitmap: str,
                 video: str,
                 wav: str) -> None:
        super().__init__()  # call base class init
        self.display_num = display_num
        self.audio_num = audio_num
        self.input = input_line
        self.output = output_line
        self.media_dir = media_dir
        self.bitmap = bitmap
        self.video = video
        self.wav = wav
        # ... anything extra here

    def fully_connected(self) -> None:
        """
        Called when the server is fully connected. Set up the "task".
        """
        print("SENDING SOME TEST/DEMONSTRATION COMMANDS")
        self.whisker.get_network_latency_ms()
        self.whisker.report_name("Whisker Twisted client prototype")
        self.whisker.timestamps(True)
        self.whisker.timer_set_event("TimerFired", 1000, 9)
        self.whisker.timer_set_event("EndOfTask", 12000)

        # ---------------------------------------------------------------------
        # Audio
        # ---------------------------------------------------------------------

        # ---------------------------------------------------------------------
        # Display
        # ---------------------------------------------------------------------
        bg_col = (0, 0, 100)
        pen = Pen(width=3, colour=(255, 255, 150), style=PenStyle.solid)
        brush = Brush(
            colour=(255, 0, 0), bg_colour=(0, 255, 0),
            opaque=True, style=BrushStyle.hatched,
            hatch_style=BrushHatchStyle.bdiagonal)
        self.whisker.claim_display(number=self.display_num, alias=DISPLAY)
        display_size = self.whisker.display_get_size(DISPLAY)
        log.info("display_size: {}".format(display_size))
        self.whisker.display_scale_documents(DISPLAY, True)
        self.whisker.display_create_document(DOC)
        self.whisker.display_set_background_colour(DOC, bg_col)
        self.whisker.display_blank(DISPLAY)
        self.whisker.display_create_document("junk")
        self.whisker.display_delete_document("junk")
        self.whisker.display_show_document(DISPLAY, DOC)
        with self.whisker.display_cache_wrapper(DOC):
            self.whisker.display_add_obj_text(
                DOC, "_text", (50, 50), "hello there!",
                italic=True)
            self.whisker.display_add_obj_line(
                DOC, "_line", (25, 25), (200, 200), pen)
            self.whisker.display_add_obj_arc(
                DOC, "_arc",
                Rectangle(left=100, top=100, width=200, height=200),
                (25, 25), (200, 200), pen)
            self.whisker.display_add_obj_bezier(
                DOC, "_bezier",
                (100, 100), (150, 100),
                (150, 200), (100, 200),
                pen)
            self.whisker.display_add_obj_chord(
                DOC, "_chord",
                Rectangle(left=300, top=0, width=100, height=100),
                (100, 150), (400, 175),
                pen, brush)
            self.whisker.display_add_obj_ellipse(
                DOC, "_ellipse",
                Rectangle(left=0, top=200, width=200, height=100),
                pen, brush)
            self.whisker.display_add_obj_pie(
                DOC, "_pie",
                Rectangle(left=0, top=300, width=200, height=100),
                (10, 320), (180, 380),
                pen, brush)
            self.whisker.display_add_obj_polygon(
                DOC, "_polygon",
                [(400, 200), (450, 300), (400, 400), (300, 300)],
                pen, brush, alternate=True)
            self.whisker.display_add_obj_rectangle(
                DOC, "_rectangle",
                Rectangle(left=500, top=0, width=200, height=100),
                pen, brush)
            self.whisker.display_add_obj_roundrect(
                DOC, "_roundrect",
                Rectangle(left=500, top=200, width=100, height=200),
                150, 250,
                pen, brush)
            self.whisker.display_add_obj_camcogquadpattern(
                DOC, "_camcogquad",
                (500, 400),
                10, 10,
                [1, 2, 4, 8, 16, 32, 64, 128],
                [255, 254, 253, 252, 251, 250, 249, 248],
                [1, 2, 3, 4, 5, 6, 7, 8],
                [128, 64, 32, 16, 8, 4, 2, 1],
                (255, 0, 0),
                (0, 255, 0),
                (0, 0, 255),
                (255, 0, 255),
                (100, 100, 100))
        self.whisker.display_set_event(DOC, "_polygon", "poly_touched")
        self.whisker.display_clear_event(DOC, "_polygon")
        self.whisker.display_set_event(DOC, "_camcogquad", "play")
        self.whisker.display_set_event(DOC, "_roundrect", "pause")
        self.whisker.display_set_event(DOC, "_rectangle", "stop")
        self.whisker.display_set_obj_event_transparency(DOC, "_rectangle",
                                                        False)
        self.whisker.display_bring_to_front(DOC, "_chord")
        self.whisker.display_send_to_back(DOC, "_chord")
        self.whisker.display_keyboard_events(DOC)
        self.whisker.display_event_coords(False)
        self.whisker.display_set_audio_device(DISPLAY, AUDIO)
        self.whisker.display_add_obj_video(DOC, VIDEO, (600, 0), self.video)
        self.whisker.video_set_volume(DOC, VIDEO, 90)
        self.whisker.video_timestamps(True)
        video_dur_ms = self.whisker.video_get_duration_ms(DOC, VIDEO)
        log.info("video_dur_ms: {}".format(video_dur_ms))
        video_pos_ms = self.whisker.video_get_time_ms(DOC, VIDEO)
        log.info("video_pos_ms: {}".format(video_pos_ms))

    def incoming_event(self, event: str, timestamp: int = None) -> None:
        """
        Responds to incoming events from Whisker.
        """
        print("Event: {e} (timestamp {t})".format(e=event, t=timestamp))
        if event == "EndOfTask":
            # noinspection PyUnresolvedReferences
            reactor.stop()
        elif event == "play":
            self.whisker.video_play(DOC, VIDEO)
        elif event == "pause":
            self.whisker.video_pause(DOC, VIDEO)
        elif event == "stop":
            self.whisker.video_stop(DOC, VIDEO)
        elif event == "back":
            self.whisker.video_seek_relative(DOC, VIDEO, -1000)
        elif event == "forward":
            self.whisker.video_seek_relative(DOC, VIDEO, 1000)
        elif event == "start":
            self.whisker.video_seek_absolute(DOC, VIDEO, 0)


def main() -> None:
    """
    Command-line parser.
    See ``--help`` for details.
    """
    logging.basicConfig()
    logging.getLogger("whisker").setLevel(logging.DEBUG)
    configure_logger_for_colour(logging.getLogger())  # configure root logger
    # print_report_on_all_logs()

    parser = argparse.ArgumentParser("Test Whisker raw socket client")
    parser.add_argument('--server', default='localhost',
                        help="Server (default: localhost)")
    parser.add_argument('--port', default=DEFAULT_PORT, type=int,
                        help="Port (default: {})".format(DEFAULT_PORT))
    parser.add_argument(
        '--display_num', default=DEFAULT_DISPLAY_NUM, type=int,
        help="Display number to use (default: {})".format(DEFAULT_DISPLAY_NUM))
    parser.add_argument(
        '--audio_num', default=DEFAULT_AUDIO_NUM, type=int,
        help="Audio device number to use (default: {})".format(
            DEFAULT_AUDIO_NUM))
    parser.add_argument(
        '--input', default=DEFAULT_INPUT_LINE, type=int,
        help="Input line number to use (default: {})".format(
            DEFAULT_INPUT_LINE))
    parser.add_argument(
        '--output', default=DEFAULT_OUTPUT_LINE, type=int,
        help="Output line number to use (default: {})".format(
            DEFAULT_OUTPUT_LINE))
    parser.add_argument(
        '--media_dir', default=DEFAULT_MEDIA_DIR, type=str,
        help="Media directory to use (default: {})".format(
            DEFAULT_MEDIA_DIR))
    parser.add_argument(
        '--bitmap', default=DEFAULT_BITMAP, type=str,
        help="Bitmap to use (default: {})".format(DEFAULT_BITMAP))
    parser.add_argument(
        '--video', default=DEFAULT_VIDEO, type=str,
        help="Video to use (default: {})".format(DEFAULT_VIDEO))
    parser.add_argument(
        '--wav', default=DEFAULT_WAV, type=str,
        help="WAV file to use (default: {})".format(DEFAULT_WAV))
    args = parser.parse_args()

    print("Module run explicitly. Running a Whisker test.")
    w = MyWhiskerTwistedTask(
        display_num=args.display_num,
        audio_num=args.audio_num,
        input_line=args.input,
        output_line=args.output,
        media_dir=args.media_dir,
        bitmap=args.bitmap,
        video=args.video,
        wav=args.wav,
    )
    w.connect(args.server, args.port)
    reactor.run()


if __name__ == '__main__':
    main()



#!/usr/bin/env python

"""
whisker/random.py

===============================================================================

    Copyright © 2011-2020 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of the Whisker Python client library.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Randomization functions that may be used by Whisker tasks.**

"""

from itertools import islice
import logging
import operator
import random
from typing import Any, Callable, Generator, Iterable, List, Sequence

from cardinal_pythonlib.lists import flatten_list, sort_list_by_index_list
from cardinal_pythonlib.reprfunc import auto_repr

log = logging.getLogger(__name__)

SHUFFLE_FUNC_TYPE = Callable[[Sequence[Any]], List[int]]


# =============================================================================
# Basic list functions
# =============================================================================

def last_index_of(x: List[Any], value: Any) -> int:
    """
    Gets the index of the last occurrence of ``value`` in the list ``x``.
    """
    return len(x) - 1 - x[::-1].index(value)


def get_unique_values(iterable: Iterable[Any]) -> List[Any]:
    """
    Gets the unique values of its input.
    See
    https://stackoverflow.com/questions/12897374/get-unique-values-from-a-list-in-python.

    (We don't use ``list(set(x))``, because if the elements of ``x``
    are themselves lists (perfectly common!), that gives ``TypeError:
    unhashable type: 'list'``.)
    """  # noqa
    seen = set()
    result = []
    for element in iterable:
        hashed = element
        if isinstance(element, dict):
            hashed = tuple(sorted(element.items()))
        elif isinstance(element, list):
            hashed = tuple(element)
        if hashed not in seen:
            result.append(element)
            seen.add(hashed)
    return result


def get_indexes_for_value(x: List[Any], value: Any) -> List[int]:
    """
    Returns a list of indexes of ``x`` where its value is ``value``.
    """
    return [i for i, item in enumerate(x) if item == value]


# =============================================================================
# Draw-without-replacement
# =============================================================================

def make_dwor_hat(values: Iterable[Any], multiplier: int = 1) -> List[Any]:
    """
    Makes a "hat" to draw values from. See :func:`gen_dwor`. Does not modify
    the starting list; returns a copy.
    """
    assert multiplier >= 1, "Bad DWOR multiplier"
    hat = []
    for v in values:
        for _ in range(multiplier):
            hat.append(v)
    random.shuffle(hat)  # shuffle in place
    return hat


def gen_dwor(values: Iterable[Any],
             multiplier: int = 1) -> Generator[Any, None, None]:
    """
    Generates values using a draw-without-replacement (DWOR) system.

    Args:
        values: values to generate
        multiplier: DWOR multiplier; see below.

    Yields:
        successive values

    Here's how it works.

    - Suppose ``values == [A, B, C]``.
    - We'll call :math:`n` the number of values (here, 3), and :math:`k` the
      "multiplier" parameter.
    - If you iterate through ``gen_dwor(values, multiplier=1)``, you will
      get a sequence that might look like this (with spaces added for
      clarity):

      .. code-block:: none

        CAB ABC BCA BAC BAC ACB CBA ...

      That is, individual are drawn randomly from a "hat" of size :math:`n =
      3`, containing one of each thing from ``values``. When the hat is empty,
      it is refilled with :math:`n` more.

    - If you iterate through ``gen_dwor(values, multiplier=2)``, however, you
      might get this:

      .. code-block:: none

        AACBBC CABBAC BAACCB ...

      The computer has put :math:`k` copies of each value in the hat, and then
      draws one each time at random (so the hat starts with :math:`nk` values
      in it). When the hat is exhausted, it re-populates.

    The general idea is to provide randomness, but randomness that is
    constrained to prevent unlikely but awkward sequences like

    .. code-block:: none

        AAAAAAAAAAAAAAAA ... unlikely but possible with full randomness!

    yet also have the option to avoid predictability. With :math:`k = 1`, then
    a clever subject could infer exactly what's coming up on every 
    :math:`n`\ th trial. So a low value of :math:`k` brings very few "runs" but
    some predictability; as :math:`k` approaches infinity, it's equivalent to
    full randomness; some reasonably low value of :math:`k` in between may be a
    useful experimental sweet spot.

    See also, for example:

    - http://www.whiskercontrol.com/help/FiveChoice/index.html?fivechoice_dwor.htm

    """  # noqa
    hat = []
    while True:
        if not hat:
            hat = make_dwor_hat(values, multiplier)
        value = hat.pop()
        yield value


def get_dwor_list(values: Iterable[Any],
                  length: int,
                  multiplier: int = 1) -> List[Any]:
    """
    Makes a fixed-length list via :func:`gen_dwor`.

    Args:
        values: values to pick from
        length: list length
        multiplier: DWOR multiplier

    Returns:
        list of length ``length``

    Example:

    .. code-block:: python

        from whisker.random import get_dwor_list
        values = ["a", "b", "c"]
        print(get_dwor_list(values, length=24, multiplier=1))
        print(get_dwor_list(values, length=24, multiplier=2))
        print(get_dwor_list(values, length=24, multiplier=3))

    """
    return list(islice(gen_dwor(values, multiplier), length))


# =============================================================================
# Simple randomness
# =============================================================================

def shuffle_list_slice(x: List[Any],
                       start: int = None, end: int = None) -> None:
    """
    Shuffles a segment of a list, ``x[start:end]``, in place.

    Note that ``start=None`` means "from the beginning" and ``end=None`` means
    "to the end".

    """
    # log.debug("x={}, start={}, end={}".format(x, start, end))
    copy = x[start:end]
    random.shuffle(copy)
    x[start:end] = copy


def shuffle_list_subset(x: List[Any], indexes: List[int]) -> None:
    """
    Shuffles some elements of a list (in place). The elements to interchange
    (shuffle) as specified by ``indexes``.
    """
    elements = [x[i] for i in indexes]
    random.shuffle(elements)
    for element_idx, x_idx in enumerate(indexes):
        x[x_idx] = elements[element_idx]


# =============================================================================
# Randomness within/across chunks
# =============================================================================

def shuffle_list_within_chunks(x: List[Any], chunksize: int) -> None:
    """
    Divides a list into chunks and shuffles WITHIN each chunk (in place).
    For example:

    .. code-block:: python

        x = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]
        shuffle_list_within_chunks(x, 4)

    ``x`` might now be:

    .. code-block:: none

        [4, 1, 3, 2, 7, 5, 6, 8, 9, 12, 11, 10]
         ^^^^^^^^^^  ^^^^^^^^^^  ^^^^^^^^^^^^^
    """
    starts = list(range(0, len(x), chunksize))
    # noinspection PyTypeChecker
    ends = starts[1:] + [None]
    for start, end in zip(starts, ends):
        shuffle_list_slice(x, start, end)


def shuffle_list_chunks(x: List[Any], chunksize: int) -> None:
    """
    Divides a list into chunks and shuffles the chunks themselves (in place).
    For example:

    .. code-block:: python

        x = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]
        shuffle_list_chunks(x, 4)

    ``x`` might now be

    .. code-block:: none

        [5, 6, 7, 8, 1, 2, 3, 4, 9, 10, 11, 12]
         ^^^^^^^^^^  ^^^^^^^^^^  ^^^^^^^^^^^^^

    Uses :func:`cardinal_pythonlib.lists.flatten_list` and
    :func:`cardinal_pythonlib.lists.sort_list_by_index_list`. (I say that
    mainly to test Intersphinx, when it is enabled.)
    """
    starts = list(range(0, len(x), chunksize))
    ends = starts[1:] + [len(x)]
    # Shuffle the indexes rather than the array, then we can write back
    # in place.
    chunks = []
    for start, end in zip(starts, ends):
        chunks.append(list(range(start, end)))
    random.shuffle(chunks)
    indexes = flatten_list(chunks)
    sort_list_by_index_list(x, indexes)


# =============================================================================
# Hierarchical randomness: early methods
# =============================================================================

def block_shuffle_by_item(x: List[Any],
                          indexorder: List[int],
                          start: int = None,
                          end: int = None) -> None:
    """
    **DEPRECATED:** :func:`layered_shuffle` is more powerful.

    Shuffles the list ``x[start:end]`` hierarchically, in place.

    Args:
        x: list to shuffle
        indexorder: a list of indexes of each item of ``x``
            The first index varies slowest; the last varies fastest.
        start: start index of ``x``
        end: end index of ``x``

    For example:

    .. code-block:: python

        p = list(itertools.product("ABC", "xyz", "123"))

    ``x`` is now a list of tuples looking like ``('A', 'x', '1')``.

    .. code-block:: python

        block_shuffle_by_item(p, [0, 1, 2])

    ``p`` might now look like:

    .. code-block:: none

        C z 1 } all values of "123" appear  } first "xyz" block
        C z 3 } once, but randomized        }
        C z 2 }                             }
                                            }
        C y 2 } next "123" block            }
        C y 1 }                             }
        C y 3 }                             }
                                            }
        C x 3                               }
        C x 2                               }
        C x 1                               }

        A y 3                               } second "xyz" block
        ...                                 } ...

    A clearer explanation is in :func:`block_shuffle_by_attr`.

    """
    item_idx_order = indexorder.copy()
    item_idx = item_idx_order.pop(0)

    # 1. Take copy
    sublist = x[start:end]

    # 2. Sort then shuffle in chunks (e.g. at the "ABC" level)
    sublist.sort(key=operator.itemgetter(item_idx))
    # Note below that we must convert things to tuples to put them into
    # sets; if you take a set() of lists, you get
    #   TypeError: unhashable type: 'list'
    unique_values = set(tuple(x[item_idx]) for x in sublist)
    chunks = [
        [i for i, v in enumerate(sublist) if tuple(v[item_idx]) == value]
        for value in unique_values
    ]
    random.shuffle(chunks)
    indexes = flatten_list(chunks)
    sort_list_by_index_list(sublist, indexes)

    # 3. Call recursively (e.g. at the "xyz" level next)
    if item_idx_order:  # more to do?
        starts_ends = [(min(chunk), max(chunk) + 1) for chunk in chunks]
        for s, e in starts_ends:
            block_shuffle_by_item(sublist, item_idx_order, s, e)

    # 4. Write back
    x[start:end] = sublist


def block_shuffle_by_attr(x: List[Any],
                          attrorder: List[str],
                          start: int = None,
                          end: int = None) -> None:
    """
    **DEPRECATED:** :func:`layered_shuffle` is more powerful.

    Exactly as for :func:`block_shuffle_by_item`, but by item attribute
    rather than item index number.

    For example:

    .. code-block:: python

        from collections import namedtuple
        import itertools
        from whisker.random import block_shuffle_by_attr

        p = list(itertools.product("ABC", "xyz", "123"))
        Trio = namedtuple("Trio", ["upper", "lower", "digit"])
        q = [Trio(*x) for x in p]
        block_shuffle_by_attr(q, ['upper', 'lower', 'digit'])

    ``q`` started off as:

    .. code-block:: none

        [
            Trio(upper='A', lower='x', digit='1'),
            Trio(upper='A', lower='x', digit='2'),
            Trio(upper='A', lower='x', digit='3'),
            Trio(upper='A', lower='y', digit='1'),
            Trio(upper='A', lower='y', digit='2'),
            Trio(upper='A', lower='y', digit='3'),
            Trio(upper='A', lower='z', digit='1'),
            Trio(upper='A', lower='z', digit='2'),
            Trio(upper='A', lower='z', digit='3'),
            Trio(upper='B', lower='x', digit='1'),
            Trio(upper='B', lower='x', digit='2'),
            Trio(upper='B', lower='x', digit='3'),
            Trio(upper='B', lower='y', digit='1'),
            Trio(upper='B', lower='y', digit='2'),
            Trio(upper='B', lower='y', digit='3'),
            Trio(upper='B', lower='z', digit='1'),
            Trio(upper='B', lower='z', digit='2'),
            Trio(upper='B', lower='z', digit='3'),
            Trio(upper='C', lower='x', digit='1'),
            Trio(upper='C', lower='x', digit='2'),
            Trio(upper='C', lower='x', digit='3'),
            Trio(upper='C', lower='y', digit='1'),
            Trio(upper='C', lower='y', digit='2'),
            Trio(upper='C', lower='y', digit='3'),
            Trio(upper='C', lower='z', digit='1'),
            Trio(upper='C', lower='z', digit='2'),
            Trio(upper='C', lower='z', digit='3')
        ]

    but after the shuffle ``q`` might now be:

    .. code-block:: none

        [
            Trio(upper='B', lower='z', digit='1'),
            Trio(upper='B', lower='z', digit='3'),
            Trio(upper='B', lower='z', digit='2'),
            Trio(upper='B', lower='x', digit='1'),
            Trio(upper='B', lower='x', digit='3'),
            Trio(upper='B', lower='x', digit='2'),
            Trio(upper='B', lower='y', digit='3'),
            Trio(upper='B', lower='y', digit='2'),
            Trio(upper='B', lower='y', digit='1'),
            Trio(upper='A', lower='z', digit='2'),
            Trio(upper='A', lower='z', digit='1'),
            Trio(upper='A', lower='z', digit='3'),
            Trio(upper='A', lower='x', digit='1'),
            Trio(upper='A', lower='x', digit='2'),
            Trio(upper='A', lower='x', digit='3'),
            Trio(upper='A', lower='y', digit='3'),
            Trio(upper='A', lower='y', digit='1'),
            Trio(upper='A', lower='y', digit='2'),
            Trio(upper='C', lower='x', digit='2'),
            Trio(upper='C', lower='x', digit='3'),
            Trio(upper='C', lower='x', digit='1'),
            Trio(upper='C', lower='y', digit='2'),
            Trio(upper='C', lower='y', digit='1'),
            Trio(upper='C', lower='y', digit='3'),
            Trio(upper='C', lower='z', digit='1'),
            Trio(upper='C', lower='z', digit='2'),
            Trio(upper='C', lower='z', digit='3')
        ]

    You can see that the ``A``/``B``/``C`` group has been shuffled as blocks.
    Then, within ``B``, the ``x``/``y``/``z`` groups have been shuffled (and so
    on for ``A`` and ``C``). Then, within ``B.z``, the ``1``/``2``/``3`` values
    have been shuffled (and so on).

    """
    item_attr_order = attrorder.copy()
    item_attr = item_attr_order.pop(0)
    # 1. Take copy
    sublist = x[start:end]
    # 2. Sort then shuffle in chunks
    sublist.sort(key=operator.attrgetter(item_attr))
    unique_values = set(tuple(getattr(x, item_attr)) for x in sublist)
    chunks = [
        [
            i for i, v in enumerate(sublist)
            if tuple(getattr(v, item_attr)) == value
        ]
        for value in unique_values
    ]
    random.shuffle(chunks)
    indexes = flatten_list(chunks)
    sort_list_by_index_list(sublist, indexes)
    # 3. Call recursively (e.g. at the "xyz" level next)
    if item_attr_order:  # more to do?
        starts_ends = [(min(chunk), max(chunk) + 1) for chunk in chunks]
        for s, e in starts_ends:
            block_shuffle_by_attr(sublist, item_attr_order, s, e)
    # 4. Write back
    x[start:end] = sublist


def shuffle_where_equal_by_attr(x: List[Any], attrname: str) -> None:
    """
    **DEPRECATED:** :func:`layered_shuffle` is more powerful.

    Shuffles a list ``x``, in place, where list members are equal as judged by
    the attribute ``attrname``.

    This is easiest to show by example:

    .. code-block:: python

        from collections import namedtuple
        import itertools
        from whisker.random import shuffle_where_equal_by_attr

        p = list(itertools.product("ABC", "xyz", "123"))
        Trio = namedtuple("Trio", ["upper", "lower", "digit"])
        q = [Trio(*x) for x in p]
        shuffle_where_equal_by_attr(q, 'digit')

    ``q`` started off as:

    .. code-block:: none

        [
            Trio(upper='A', lower='x', digit='1'),
            Trio(upper='A', lower='x', digit='2'),
            Trio(upper='A', lower='x', digit='3'),
            Trio(upper='A', lower='y', digit='1'),
            Trio(upper='A', lower='y', digit='2'),
            Trio(upper='A', lower='y', digit='3'),
            Trio(upper='A', lower='z', digit='1'),
            Trio(upper='A', lower='z', digit='2'),
            Trio(upper='A', lower='z', digit='3'),
            Trio(upper='B', lower='x', digit='1'),
            Trio(upper='B', lower='x', digit='2'),
            Trio(upper='B', lower='x', digit='3'),
            Trio(upper='B', lower='y', digit='1'),
            Trio(upper='B', lower='y', digit='2'),
            Trio(upper='B', lower='y', digit='3'),
            Trio(upper='B', lower='z', digit='1'),
            Trio(upper='B', lower='z', digit='2'),
            Trio(upper='B', lower='z', digit='3'),
            Trio(upper='C', lower='x', digit='1'),
            Trio(upper='C', lower='x', digit='2'),
            Trio(upper='C', lower='x', digit='3'),
            Trio(upper='C', lower='y', digit='1'),
            Trio(upper='C', lower='y', digit='2'),
            Trio(upper='C', lower='y', digit='3'),
            Trio(upper='C', lower='z', digit='1'),
            Trio(upper='C', lower='z', digit='2'),
            Trio(upper='C', lower='z', digit='3')
        ]

    but after the shuffle ``q`` might now be:

    .. code-block:: none

        [
            Trio(upper='A', lower='x', digit='1'),
            Trio(upper='A', lower='y', digit='2'),
            Trio(upper='A', lower='z', digit='3'),
            Trio(upper='B', lower='z', digit='1'),
            Trio(upper='A', lower='z', digit='2'),
            Trio(upper='C', lower='x', digit='3'),
            Trio(upper='B', lower='y', digit='1'),
            Trio(upper='A', lower='x', digit='2'),
            Trio(upper='C', lower='y', digit='3'),
            Trio(upper='A', lower='y', digit='1'),
            Trio(upper='C', lower='y', digit='2'),
            Trio(upper='C', lower='z', digit='3'),
            Trio(upper='C', lower='y', digit='1'),
            Trio(upper='C', lower='z', digit='2'),
            Trio(upper='A', lower='y', digit='3'),
            Trio(upper='B', lower='x', digit='1'),
            Trio(upper='B', lower='z', digit='2'),
            Trio(upper='B', lower='y', digit='3'),
            Trio(upper='C', lower='z', digit='1'),
            Trio(upper='C', lower='x', digit='2'),
            Trio(upper='B', lower='z', digit='3'),
            Trio(upper='C', lower='x', digit='1'),
            Trio(upper='B', lower='x', digit='2'),
            Trio(upper='A', lower='x', digit='3'),
            Trio(upper='A', lower='z', digit='1'),
            Trio(upper='B', lower='y', digit='2'),
            Trio(upper='B', lower='x', digit='3')
        ]

    As you can see, the ``digit`` attribute seems to have stayed frozen and
    everything else has jumbled. What has actually happened is that everything
    with ``digit == 1`` has been shuffled among themselves, and similarly for
    ``digit == 2`` and ``digit == 3``.

    """
    unique_values = set(tuple(getattr(item, attrname)) for item in x)
    for value in unique_values:
        indexes = [
            i for i, item in enumerate(x)
            if tuple(getattr(item, attrname)) == value
        ]
        shuffle_list_subset(x, indexes)


# =============================================================================
# Hierarchical randomness: full power
# =============================================================================

def random_shuffle_indexes(x: List[Any]) -> List[int]:
    """
    Returns a list of indexes of ``x``, randomly shuffled.
    """
    indexes = list(range(len(x)))
    random.shuffle(indexes)  # in place
    return indexes


def block_shuffle_indexes_by_value(x: List[Any]) -> List[int]:
    """
    Returns a list of indexes of ``x``, block-shuffled by value.

    That is: we aggregate items into blocks, defined by value, and shuffle
    those blocks, returning the corresponding indexes of the original list.
    """
    unique_values = get_unique_values(x)
    list_of_chunks = [
        get_indexes_for_value(x, value)
        for value in unique_values
    ]
    random.shuffle(list_of_chunks)
    indexes = flatten_list(list_of_chunks)
    return indexes


def dwor_shuffle_indexes(x: List[Any], multiplier: int = 1) -> List[int]:
    """
    Returns a list of indexes of ``x``, DWOR-shuffled by value.

    This is a bit tricky as we don't have a guarantee of equal numbers. It
    does sensible things in those circumstances.
    """
    assert multiplier >= 1, "Bad DWOR multiplier"
    unique_values = get_unique_values(x)
    list_of_chunks = [
        get_indexes_for_value(x, value)
        for value in unique_values
    ]

    def unusual_make_dwor_hat() -> List[int]:
        hat_ = []
        for _ in range(multiplier):
            for chunk in list_of_chunks:
                if chunk:
                    hat_.append(chunk.pop(0))
        random.shuffle(hat_)  # in place
        return hat_

    hat = []
    indexes = []  # type: List[int]
    finished = False
    while not finished:
        if not hat:
            hat = unusual_make_dwor_hat()
        if hat:
            indexes.append(hat.pop())
        else:
            finished = True
    return indexes


def sort_indexes(x: List[Any]) -> List[int]:
    """
    Returns the indexes of ``x`` in an order that would sort ``x`` by value.
    
    See https://stackoverflow.com/questions/7851077/how-to-return-index-of-a-sorted-list
    """  # noqa
    return sorted(range(len(x)), key=lambda index: x[index])


def reverse_sort_indexes(x: List[Any]) -> List[int]:
    """
    Returns the indexes of ``x`` in an order that would reverse-sort ``x`` by
    value.
    """
    return sorted(range(len(x)), key=lambda index: x[index], reverse=True)


class ShuffleLayerMethod(object):
    """
    Class to representing instructions to :func:`layered_shuffle` (q.v.).
    """
    def __init__(self,
                 flat: bool = False,
                 layer_key: int = None,
                 layer_attr: str = None,
                 layer_func: Callable[[Any], Any] = None,
                 shuffle_func: SHUFFLE_FUNC_TYPE = None) -> None:
        """
        Args:
            flat: take data as ``x[index]``?
            layer_key: take data as ``x[index][layer_key]``?
            layer_attr: take data as ``getattr(x[index], layer_attr)``?
            layer_func: take data as ``layer_func(x[index])``?
            shuffle_func: function (N.B. may be a lambda function with
                parameters attached) that takes a list of objects and returns
                a list of INDEXES, suitably shuffled.

        Typical values of ``shuffle_func``:

        - ``None``: no shuffle
        - ``random_shuffle_indexes``: plain shuffle; see
          :func:`random_shuffle_indexes`
        - ``dwor_shuffle_indexes``: DWOR-style shuffle (will need a ``lambda``
          for its ``multiplier parameter``; see :func:`dwor_shuffle_indexes`
        - ``block_shuffle_indexes_by_value``: aggregate items into blocks,
          defined by value, and shuffle those blocks; see
          :func:`block_shuffle_indexes_by_value`
        - ``sort_indexes``: not exactly shuffling! See :func:`sort_indexes`.
        - ``reverse_sort_indexes``: not exactly shuffling! See
          :func:`reverse_sort_indexes`.

        """
        assert (
            flat +
            (layer_key is not None) +
            bool(layer_attr)
        ) == 1, "Specify exactly one way of accessing data from the layer!"
        if layer_key is not None:
            assert isinstance(layer_key, int)
            assert layer_key > 0
        elif layer_attr:
            assert isinstance(layer_attr, str)
        elif layer_func:
            assert callable(layer_func)
        self.flat = flat
        self.layer_key = layer_key
        self.layer_attr = layer_attr
        self.layer_func = layer_func
        self.shuffle_func = shuffle_func

    def __repr__(self) -> str:
        return auto_repr(self)

    def get_values(self, x: List[Any]) -> List[Any]:
        """
        Returns all the values of interest from ``x`` for this layer.
        """
        if self.flat:
            return x[:]  # a copy, just in case someone wants to modify it
        elif self.layer_attr:
            attr = self.layer_attr
            return [getattr(item, attr) for item in x]
        elif self.layer_func:
            func = self.layer_func
            return [func(item) for item in x]
        else:
            # key, via item[key], meaning item.__getitem__(key)
            key = self.layer_key
            return [item[key] for item in x]

    def get_unique_values(self, x: List[Any]) -> List[Any]:
        """
        Returns all the unique values of ``x`` for this layer.
        """
        return get_unique_values(self.get_values(x))

    def get_indexes_for_value(self, x: List[Any], value: Any) -> List[int]:
        """
        Returns a list of indexes of ``x`` where its value (as defined by this
        layer) is ``value``.
        """
        return get_indexes_for_value(self.get_values(x), value)


def layered_shuffle(x: List[Any],
                    layers: List[ShuffleLayerMethod]) -> None:
    r"""
    Most powerful hierarchical shuffle command here.

    Shuffles ``x`` in place in a layered way as specified by the sequence of
    methods.
    
    In more detail:
    
    - for each layer, it shuffles values of ``x`` as defined by the
      :class:`ShuffleLayerMethod` (for example: "shuffle ``x`` in blocks based
      on the value of ``x.someattr``", or "shuffle ``x`` randomly")
    - it then proceeds to deeper layers *within* sub-lists defined by each
      unique value from the previous layer. 

    Args:
        x: sequence (e.g. list) to shuffle
        layers: list of :class:`ShuffleLayerMethod` instructions

    Examples:

    .. code-block:: python

        from collections import namedtuple
        import itertools
        import logging
        import random
        from whisker.random import *
        logging.basicConfig(level=logging.DEBUG)
        
        startlist = ["a", "b", "c", "d", "a", "b", "c", "d", "a", "b", "c", "d"]
        x1 = startlist[:]
        x2 = startlist[:]
        x3 = startlist[:]
        x4 = startlist[:]
        
        do_nothing_method = ShuffleLayerMethod(flat=True, shuffle_func=None)
        do_nothing_method.get_unique_values(x1)
        do_nothing_method.get_indexes_for_value(x1, "b")
        
        layered_shuffle(x1, [do_nothing_method])
        print(x1)
        
        flat_randomshuffle_method = ShuffleLayerMethod(
            flat=True, shuffle_func=random_shuffle_indexes)
        flat_randomshuffle_method.get_unique_values(x1)
        flat_randomshuffle_method.get_indexes_for_value(x1, "b")
        layered_shuffle(x1, [flat_randomshuffle_method])
        print(x1)
        
        flat_blockshuffle_method = ShuffleLayerMethod(
            flat=True, shuffle_func=block_shuffle_indexes_by_value)
        layered_shuffle(x2, [flat_blockshuffle_method])
        print(x2)
        
        flat_dworshuffle_method = ShuffleLayerMethod(
            flat=True, shuffle_func=dwor_shuffle_indexes)
        layered_shuffle(x3, [flat_dworshuffle_method])
        print(x3)
        
        flat_dworshuffle2_method = ShuffleLayerMethod(
            flat=True, shuffle_func=lambda x: dwor_shuffle_indexes(x, multiplier=2))
        layered_shuffle(x4, [flat_dworshuffle2_method])
        print(x4)
        
        p = list(itertools.product("ABC", "xyz", "123"))
        Trio = namedtuple("Trio", ["upper", "lower", "digit"])
        q = [Trio(*x) for x in p]
        print("\n".join(str(x) for x in q))
        
        upper_method = ShuffleLayerMethod(
            layer_attr="upper", shuffle_func=block_shuffle_indexes_by_value)
        lower_method = ShuffleLayerMethod(
            layer_attr="lower", shuffle_func=reverse_sort_indexes)
        digit_method = ShuffleLayerMethod(
            layer_attr="digit", shuffle_func=random_shuffle_indexes)
        
        layered_shuffle(q, [upper_method, lower_method, digit_method])
        print("\n".join(str(x) for x in q))

    """  # noqa
    if not layers:
        return
    layer = layers[0]
    subsequent_layers = layers[1:]

    if layer.shuffle_func:
        values = layer.get_values(x)
        shuffled_indexes = layer.shuffle_func(values)
        log.debug("Applying function {!r} to values {!r} gives indexes "
                  "{!r}".format(layer.shuffle_func, values, shuffled_indexes))
        sort_list_by_index_list(x, shuffled_indexes)

    if subsequent_layers:
        unique_values = layer.get_unique_values(x)
        for value in unique_values:
            indexes_for_value = layer.get_indexes_for_value(x, value)
            subelements = [x[i] for i in indexes_for_value]
            # Recursion:
            layered_shuffle(subelements, subsequent_layers)
            # Put the (shuffled( elements back:
            for element_idx, x_idx in enumerate(indexes_for_value):
                x[x_idx] = subelements[element_idx]



#!/usr/bin/env python

"""
whisker/debug_qt.py

===============================================================================

    Copyright © 2011-2020 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of the Whisker Python client library.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Functions to debug Qt objects and signals under PyQt5.**

- See http://stackoverflow.com/questions/2045352

"""


import logging
import threading
from typing import Any, Callable

# noinspection PyPackageRequirements
from PyQt5.QtCore import QObject, QThread
from PyQt5.QtCore import pyqtBoundSignal

log = logging.getLogger(__name__)

TYPE_CONNECT_FN = Callable[[Any], None]
TYPE_DISCONNECT_FN = Callable[[Any], None]
TYPE_EMIT_FN = Callable[[Any], None]

_old_connect = pyqtBoundSignal.connect  # normal method
_old_disconnect = pyqtBoundSignal.disconnect  # normal method
_old_emit = pyqtBoundSignal.emit  # normal method


def _wrap_connect(callable_object):
    """
    Returns a wrapped call to the old version of
    :func:`pyqtBoundSignal.connect`.
    """

    def call(self, *args, **kwargs) -> None:
        callable_object(self, *args, **kwargs)
        _old_connect(self, *args, **kwargs)

    return call


def _wrap_disconnect(callable_object):
    """
    Returns a wrapped call to the old version of
    :func:`pyqtBoundSignal.disconnect`.
    """

    def call(self, *args, **kwargs) -> None:
        callable_object(self, *args, **kwargs)
        _old_disconnect(self, *args, **kwargs)

    return call


def _wrap_emit(callable_object):
    """
    Returns a wrapped call to the old version of
    :func:`pyqtBoundSignal.emit`.
    """

    def call(self, *args) -> None:
        callable_object(self, *args)
        _old_emit(self, *args)

    return call


def enable_signal_debugging(connect_call: TYPE_CONNECT_FN = None,
                            disconnect_call: TYPE_DISCONNECT_FN = None,
                            emit_call: TYPE_EMIT_FN = None) -> None:
    """
    Call this to enable PySide/Qt signal debugging. This will trap all
    :func:`connect` and :func:`disconnect` calls, calling the user-supplied
    functions first and then the real things.
    """

    # noinspection PyUnusedLocal
    def cd(*args, **kwargs) -> None:
        pass

    # noinspection PyUnusedLocal
    def e(*args) -> None:
        pass

    connect_call = connect_call or cd
    disconnect_call = disconnect_call or cd
    emit_call = emit_call or e

    pyqtBoundSignal.connect = _wrap_connect(connect_call)
    pyqtBoundSignal.disconnect = _wrap_disconnect(disconnect_call)
    pyqtBoundSignal.emit = _wrap_emit(emit_call)


def simple_connect_debugger(*args) -> None:
    """
    Function to report on :func:`connect` calls.
    """
    log.debug("CONNECT: signal={s}, args={a}".format(
        s=args[0],
        a=args[1:]
    ))


def simple_emit_debugger(*args) -> None:
    """
    Function to report on :func:`emit` calls.
    """
    emitter = args[0]
    # emitter_qthread = emitter.thread()
    log.debug(
        "EMIT: emitter={e}, "
        "thread name={n}, signal={s}, args={a}".format(
            e=emitter,
            n=threading.current_thread().name,
            s=repr(args[1]),
            a=repr(args[2:]),
        )
        # emitter's thread={t}, currentThreadId={i}, "
        # t=emitter_qthread,
        # i=emitter_qthread.currentThreadId(),
    )


def enable_signal_debugging_simply() -> None:
    """
    Enables Qt signal debugging for :func:`connect` and :func:`emit` calls.
    """
    enable_signal_debugging(connect_call=simple_connect_debugger,
                            emit_call=simple_emit_debugger)


def debug_object(obj: QObject) -> None:
    """
    Writes a debug log message within information about the QObject.
    """
    log.debug("Object {} belongs to QThread {}".format(obj, obj.thread()))
    # Does nothing if library compiled in release mode:
    # log.debug("... dumpObjectInfo: {}".format(obj.dumpObjectInfo()))
    # log.debug("... dumpObjectTree: {}".format(obj.dumpObjectTree()))


def debug_thread(thread: QThread) -> None:
    """
    Writes a debug log message about the QThread.
    """
    log.debug("QThread {}".format(thread))



#!/usr/bin/env python

"""
whisker/callback.py

===============================================================================

    Copyright © 2011-2020 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of the Whisker Python client library.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Classes to implement callbacks via the Whisker API.**

"""

import logging
from typing import Any, Callable, Dict, List, Tuple

log = logging.getLogger(__name__)


# =============================================================================
# CallbackDefinition
# =============================================================================

class CallbackDefinition(object):
    """
    Event callback handler. Distinct from any particular network/threading
    model, so all can use it.
    """
    def __init__(self,
                 event: str,
                 callback: Callable[..., Any],
                 args: List[Any] = None,
                 kwargs: Dict[str, Any] = None,
                 target_n_calls: int = 0,
                 swallow_event: bool = False) -> None:
        """
        Args:
            event: Whisker event name
            callback: user-supplied callback function
            args: positional arguments to ``callback``
            kwargs: keyword arguments to ``callback``
            target_n_calls: ``0`` to keep calling indefinitely, os a positive
                integer to make that many calls (upon receiving the relevant
                event) but then to forget about the callback and mark it as
                defunct
            swallow_event: make the API swallow the event, so it's not seen by
                our client task?
        """
        self.event = event
        self.callback = callback
        self.args = args or []  # type: List[Any]
        self.kwargs = kwargs or {}  # type: Dict[str, Any]
        self.target_n_calls = target_n_calls
        self.swallow_event = swallow_event
        self.n_calls = 0

    def __repr__(self) -> str:
        return (
            "CallbackDefinition(event={}, callback={}, args={}, "
            "kwargs={}, target_n_calls={}, n_calls={}".format(
                self.event, repr(self.callback), repr(self.args),
                repr(self.kwargs), self.target_n_calls, self.n_calls,
            )
        )

    def call(self) -> None:
        """
        Calls the callback function.
        """
        self.n_calls += 1
        log.debug(
            "Callback #{n_calls} to {func}, args={args}, "
            "kwargs={kwargs}".format(
                n_calls=self.n_calls,
                func=self.callback.__name__,
                args=self.args,
                kwargs=self.kwargs,
            )
        )
        self.callback(*self.args, **self.kwargs)

    def is_defunct(self) -> bool:
        """
        Is this callback defunct, by virtue of having been called already as
        many times as the user asked?
        """
        return 0 < self.target_n_calls <= self.n_calls


class CallbackHandler(object):
    """
    Implements callbacks based on Whisker events.
    """

    def __init__(self) -> None:
        self.callbacks = []  # list of WhiskerCallbackDefinition objects

    def add(self,
            target_n_calls: int,
            event: str,
            callback: Callable[..., Any],
            args: List[Any] = None,
            kwargs: Dict[str, Any] = None,
            swallow_event: bool = True) -> None:
        """
        Adds a callback to the handler.

        See :class:`CallbackDefinition`.
        """
        cd = CallbackDefinition(event, callback, args, kwargs,
                                target_n_calls=target_n_calls,
                                swallow_event=swallow_event)
        self.callbacks.append(cd)

    def add_single(self,
                   event: str,
                   callback: Callable[..., Any],
                   args: List[Any] = None,
                   kwargs: Dict[str, Any] = None,
                   swallow_event: bool = True) -> None:
        """
        Adds a single-shot callback for the specified event.
        """
        self.add(1, event, callback, args, kwargs, swallow_event=swallow_event)

    def add_persistent(self,
                       event: str,
                       callback: Callable[..., Any],
                       args: List[Any] = None,
                       kwargs: Dict[str, Any] = None,
                       swallow_event: bool = True) -> None:
        """
        Adds a persistent callback for the specified event.
        """
        self.add(0, event, callback, args, kwargs, swallow_event=swallow_event)

    def remove(self, event: str, callback: Callable[..., None] = None) -> None:
        """
        Removes a callback (either by event/callback pair, or all callbacks
        for an event.
        """
        self.callbacks = [
            x for x in self.callbacks
            if not (x.event == event and (callback is None or
                                          x.callback == callback))]

    def clear(self) -> None:
        """
        Removes all callbacks.
        """
        self.callbacks = []

    def process_event(self, event: str) -> Tuple[int, bool]:
        """
        Calls any callbacks registered for the event. Returns the number of
        callbacks called.
        """
        n_called = 0
        swallow_event = False
        for x in self.callbacks:
            if x.event == event:
                x.call()
                swallow_event = swallow_event or x.swallow_event
        # Remove any single-shot callbacks
        self.callbacks = [x for x in self.callbacks if not x.is_defunct()]
        return n_called, swallow_event

    def debug(self) -> None:
        """
        Write a description of our callbacks to the debug log.
        """
        log.debug("CallbackHandler: callbacks are...")
        for c in self.callbacks:
            log.debug(repr(c))
        log.debug("... end of CallbackHandler calls")
